#####################################################################
#:TriggerCacheMgr.pm
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2006, 2008. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#

=head1 NAME

CQCC::TriggerCacheMgr - Manage the central cache directory 

=head1 CLASS DESCRIPTION

This class manages central caching in the Base CC/CQ integration and 
will do the following:

1. Recognize and safely post unposted CQ transaction batches to
   ensure data integrity

2. Clean out stale files such as cqsessions and series caches
   accumulated in general use

3. Recognize the presence of error logs and optionally show the
   highlights of each log, i.e. the basic error messages

4. Clean out stale error and debug log files left over from diagnostics 

5. Delete stale session, host, and user directories over a certain
   age. This is more important with CCRC based session files since
   these have a shorter lifespan and are more likely to never be
   reused after the client session is completed.

6. Provide the administrator with a statistical summary of cache mgmt 
   activity 

7. Alert the administrator via email to completed or error-prone runs 

It is launched using the cqcc_cache_mgr script either as a ClearCase
Scheduler task or from the command line. See the help information for
cqcc_cache_mgr.bat or .pl for more information (cqcc_cache_mgr -help).

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package CQCC::TriggerCacheMgr
######################################################################

package CQCC::TriggerCacheMgr;
use CQCC::TriggerCQCC;
@ISA = ('CQCC::TriggerCQCC');
use strict;

use CQCC::IOBase;
use CQCC::Parm;
use IO::File;
use File::Basename;
require Config;

sub Version { return '2.0' };

######################################################################
#::Class Globals
######################################################################

my $windows = ($Config::Config{osname} eq "MSWin32") ? 1 : 0;
my $ps = ($windows ? "\\" : "/");  #Path Separator

########################################
# Output management: 
# This trigger is only intended to be run from the command line 
# without any user input being provided. The IOBase class is
# forced to be the base IO class with no GUI interaction enabled.
# CQ Login must either be cached or provided using a command line
# option with an encrypted password. 

&ForceParm("IOBaseInterface","CQCC::IOBase");
&ForceParm("CQCC_GUI_ENABLE","OFF");

my $outputLevel = 0;   # 0 = errors-only, 1 = dir-stats, 2 = file ops
                       # 3 = verbose debugging

########################################
# Statistics are accumulated for each session directory and
# then added to overall total statistics. 

my %sessionStats = (
    "DeletedFiles" => 0,                    
    "DeletedSessionDirs" => 0,
    "FoundErrorLogs" => 0,
    "FoundOutputLogs" => 0,
    "FoundCQBatch" => 0,
    "FoundCQBatchStale" => 0,
    "PostCQBatchStaleOK" => 0,
    "PostCQBatchStaleFail" => 0,
);

my %stats = (
    "DeletedHostDirs" => 0,
    "DeletedUserDirs" => 0,
    "TotalSessions" => 0,
    "TotalUserDirs" => 0,
);

# Create overall stats to match sessionStats
for (sort keys %sessionStats) {
    $stats{$_}=0;
}

######################################################################
#::function ClassInit

=over 6

=item C<ClassInit()>

This class function is automatically invoked to define class
parameters using CQCC::Parm objects. 

Note: Some CM parameters are defined in IOBase for access to
the main runtime trigger in doing self management. 

Current parameters include:

=over 4

=item C<CM_PATH>

Name of specific session pathname to manage, relative to the 
cache root directory

=item C<CM_USER>

Name of specific user subtree to manage.

=item C<CM_CQLOGIN>

String providing CQ login information in the form "name;encryptedPassword"

=item C<CM_DIR_STALE_HRS>

Number of hours before an unused cache subdirectory is considered abandoned
and available to be deleted. 

=item C<CM_NOEXEC>

Runtime option to describe what the cache manager would do without 
actually doing it, either in terms of deleting files or posting transaction
batches. 

=item C<CM_VERBOSE>

Level of verbose output ranging from 0 (minimal) to 3 (maximum). 

=back

=back

=cut

######################################################################
sub ClassInit {
    # Environment Parameters

    if (&ParmDefined("CM_PATH")) {
        return;
    }

    # Configuration Parameters

    &SetParmDefaults(PCLASS=>"CQCC::TriggerCacheMgr");

    &DefineParm(NAME=>"CM_PATH", DEFAULT=>"");

    &DefineParm(NAME=>"CM_USER", DEFAULT=>"");
 
    &DefineParm(NAME=>"CM_CQLOGIN", DEFAULT=>"");

    &DefineParm(NAME=>"CM_CQBATCH_POST_MINS", DEFAULT=>60,
                RANGE_MIN=>3, RANGE_MAX=>6000);

    &DefineParm(NAME=>"CM_DIR_STALE_HRS", DEFAULT=>7*24,
                RANGE_MIN=>8, RANGE_MAX=>1000);

    &DefineParm(NAME=>"CM_NOEXEC", DEFAULT=>"0",
                RANGE_LIST=>["0","1"]);

    &DefineParm(NAME=>"CM_VERBOSE", DEFAULT=>"2",
                RANGE_MIN=>0, RANGE_MAX=>3);

    &SetParmDefaults();
}

&ClassInit();

######################################################################
#::method MenuBuild

=over 6

=item C<MenuBuild()>

This method is currently used to dummy out the main menu. It may
later be modified to provide an interactive menu for this class. 

=back

=cut

######################################################################

sub MenuBuild {
    my ($self,$menuName) = @_;
}

######################################################################
#::method new

=over 6

=item C<new CQCC::TriggerCacheMgr()>

This method makes a new instance of the TriggerCacheMgr class.
See the main introduction for how to launch this class using
cqcc_cache_mgr. 

=back

=cut

######################################################################

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;  # Object or class name
    my $self = new CQCC::TriggerCQCC();   #Make instance of superclass

    # Attributes
    $self->{NAME} = "CQCC::TriggerCacheMgr";

    # This will process any remaining arguments and add/override
    # the existing instance vars.

    my @args = @_;
    while (@args) {
        my $key = shift @args;
        $self->{$key} = shift @args;
    }

    # This will turn the $self hash into an object instance.
    bless($self, $class);
   
    my $time = localtime(time);

    if (! defined $ENV{"CQCC_CACHE_MGR_SUBPROCESS"}) {
        &logMessage(0,
            "\n::###################################" . 
            "###################################\n" . 
            msg(["TCM_T_LOG_HEADER",$time,&GetParm("CQCC_CACHE_ROOT")]) .
            "\n\n");
    }

    return $self;
}

######################################################################
#::method Op_manage

=over 6

=item C<Op_manage()>

This method manages the central cache directory based on provided 
options to report on or actively clean the directories in the tree
structure. 

It takes a comma separate list of name=value pairs to override default
values of parameters that control its behavior. 

Parameters:
* name=value comma separated list
* post against Parms - easier to screen for acceptable values, debug, etc. 

Stats
* counts
* specific name lists - users, user-hosts, user-hosts[sessions]

=back

=cut

######################################################################

sub Op_manage {
    my ($self,@parmPairs) = @_;

    # (1) Process parameters from command line on top of 
    # other config parameters determine basic caching dir info
    # Exit if not using central directory

    for (@parmPairs) {
        my ($name,$value) = split("=",$_);
        &ForceParm($name,$value);
    }
    $outputLevel = &GetParm("CM_VERBOSE");
    my $cacheDir = &GetParm("CQCC_CACHE_ROOT");
    if ($cacheDir eq "") {
        &IOMessage("FATAL", ["TCM_E_CACHE_ACCESS"]);
    }

    processCQlogin();

    # (2) Walk the tree starting at the top assuming root/user/host/session
    # hierarchy, processing each session node in the tree

    my $walkResults = 0;

    # CM_PATH takes an explicit cache directory subtree path including
    # the session name.

    if (&GetParm("CM_PATH") ne "") {
        # processNode
        my $fullPath = join($ps,$cacheDir,&GetParm("CM_PATH"));
        if (-d $fullPath) {
            my $result = $self->processSession($cacheDir,&GetParm("CM_PATH"));
            if ($result != 0) {
                $walkResults++;
            }
        } else { 
            &IOMessage("FATAL", ["TCM_E_NONEXIST_DIR", $fullPath]);
        }
    } else {
        # walk user directories, add to stats
        my @userDirs = dirList($cacheDir);
        my $userDir;

        # CM_USER specifies a particular user's subtree to manage

        if (&GetParm("CM_USER") ne "") {
            @userDirs = ();
            my $userDir = &GetParm("CM_USER");
            if (-d join($ps,$cacheDir,$userDir)) {
                push @userDirs,$userDir;
            }
        }
        for $userDir (@userDirs) {
            # walk host directories, add to stats
            my @hostDirs = dirList($cacheDir, $userDir);
            my $hostDir; 
            $stats{"TotalUserDirs"}++;
            for $hostDir (@hostDirs) {
                # walk session directories, add to stats
                my $hostDirName = join($ps,$cacheDir,$userDir,$hostDir);
                my @sessionDirs = dirList($hostDirName);
                my $sessionDir; 
                for $sessionDir (@sessionDirs) {
                    my $sessionPath = join($ps,$userDir,$hostDir,$sessionDir);
                    my $result = $self->processSession($cacheDir,$sessionPath);

                    if ($result != 0) {
                        $walkResults++;
                    }
                }
                # Delete empty host directories
                @sessionDirs = dirList($hostDirName);
                if ($#sessionDirs == -1) {
                    my ($status,$output) = 
                        &deleteDir($hostDirName,"","DeletedHostDirs",
                                   "RmDirIfEmpty");
                    my $shortDirName = join($ps,$userDir,$hostDir);
                    logMessage(1,"\n$shortDirName: $output");
                }
            }
            # Delete empty user directories
            @hostDirs = dirList($cacheDir, $userDir);            
            if ($#hostDirs == -1) {
                my $userDirName = join($ps,$cacheDir,$userDir);
                my ($status,$output) = 
                     &deleteDir($userDirName,"","DeletedUserDirs",
                                "RmDirIfEmpty");
                logMessage(1,"\n$userDir: $output");
            }
        }
    }

    # Print statistics

    logMessage(0,"\n" . msg(["TCM_T_RUN_TOTALS"]) . "\n");
    my $line = sprintf("%-25s  %-10s", 
                       msg(["TCM_T_VALUE"]), msg(["TCM_T_COUNT"]));
    logMessage(0, "$line\n");
    for (sort keys %stats) {
        if ($stats{$_} > 0) {
            $line = sprintf("%-25s  %10d", $_,$stats{$_});
            logMessage(0, "$line\n");
        }
    }
    logMessage(0, "Errors: $walkResults\n");

    return $walkResults;
}

######################################################################
#::method processSession

=over 6

=item C<processSession(cacheDir, sessionDirPath)>

This method manages individual session subdirectories and does the 
following: 

* Delete stale files: Delete transient files over an established age
  (as defined by the existing CQCC_CACHE_MGMT config parameter). These
  include the series cache files (cqcc_cache.* and .cqsession* files)

* Post batches: Post stale ClearQuest transaction batches
  (.cqcc_assoc_batch) over a given age.

* Identify and manage recent error and debug logs 

* Report on the existence of error logs in the session directories
     younger than a specified time and delete those over a given age.

* Report on the existing of debug output logs over a given age and
     delete those over a given age.

* Delete stale directories: If directories have no significant
  contents after being cleaned and have seen no file changes within a
  specified period, then they should be deleted. This is particularly
  true of CCRC based session directories since the user will not
  generally be accessing ths same directory again (until they are
  assigned the same session id over time). If session directories only
  contain session wide preferences (cqcc_cache), successfully posted
  batches, and log and error files over the specified age, the entire
  directory should be deleted. Likewise empty user-host and user
  directories should be deleted.

=back

=cut

######################################################################

sub processSession {
    my ($self,$cacheDir,$sessionDir) = @_;
    my $dir = join($ps,$cacheDir,$sessionDir);
    my @files = dirList($dir);
    my $file;
    my $batchFile;

    #  Stale - cqsessions and cqcc_cache files
    #  DirStale - files and dirs deleted if over this age
    #  CQBatchStale - when to post batches

    my $maxAgeHours = &GetParm("CM_STALE_HRS");
    my $maxAgeSecs = $maxAgeHours * 60 * 60;

    my $maxCQAgeMins = &GetParm("CM_CQSESSION_STALE_MINS");
    my $maxCQAgeSecs = $maxCQAgeMins * 60;

    my $maxDirAgeHours = &GetParm("CM_DIR_STALE_HRS");
    my $maxDirAgeSecs = $maxDirAgeHours * 60 * 60;

    my @ReasonsNotToRmDir = ();
    my $postOverAge = &GetParm("CM_CQBATCH_POST_MINS");
    my $postOverAgeSecs = $postOverAge * 60;
    my $currentTime = time();
    my $batchBaseName = basename($self->{BATCHFILE});
    my $unpostedCQbatch = 0;
    my $postCQbatch = 0;
    my $youngestAge = 0;
    my $tab1 = "    ";
    my $tab2 = $tab1 . $tab1;
    my $tab3 = $tab1 . $tab2;
    my $errorCount = 0;
    my $dirBuffer = "";  # Accumulated output for dir

    ########################################   
    # Initialize Session Statistics 

    for (sort keys %sessionStats) {
        $sessionStats{$_}=0;
    }

    ########################################
    # Pattern Control Table: This table is used
    # by processSession to determine what action to
    # take regarding different file types. 
    # Note: Don't depend on sort order of patterns  
    
    my %filePatterns = (
        "\\.ccase_cache\\.\(\\S*\)" => 
            "Rule=CCaseCacheFiles,RmIfStale,RmIfDirStale",
        "\\.cqcc_clean\\.time" => 
            "Rule=CacheCleanFile,RmIfDirStale",
        "\\.cqparams" => 
            "Rule=CQPasswordFile,RmIfDirStale",
        "cqcc_cache\\.\(\\S*\)" => 
            "Rule=CacheFiles,RmIfStale,RmIfDirStale",
        "cqcc_cache\$" => 
            "Rule=SessionCacheFiles,RmIfDirStale",
        "cqsession\\.\(\\S*\)" => 
            "Rule=CQSession,RmIfStale,RmIfDirStale",
        "error\\.(\\S*)log" =>
            "Rule=ErrorLog,Count=FoundErrorLogs,RmIfDirStale,Report",
        "output\\.(\\S*)log" =>
            "Rule=OutputLog,Count=FoundOutputLogs,RmIfDirStale,Report",
        "\\.LOCK" =>
            "Rule=OperationLock,RmIfStale,RmIfDirStale",
        "$batchBaseName\\.old" =>
            "Rule=CQBatchFileOld,RmIfDirStale",
        "$batchBaseName\$" =>
            "Rule=CQBatchFile,Count=FoundCQBatch,PostCQ",
        );

    for $file (@files) {
        ########################################
        # Initialize per-file information

        my $fileBuffer = ""; # Accumulated messages for file
        my $path = $dir . $ps . $file;
        my @fileInfo = stat($path);
        my $fileSize = $fileInfo[7];
        my $fileAccessTime = $fileInfo[9];
        my $fileModTime = $fileInfo[9];
        my $age = $currentTime - $fileModTime;
        if ($youngestAge == 0 || $age < $youngestAge) {
            $youngestAge = $age;
        }
        my $deleteFile = "";

        ########################################
        # Look up file against patterns

        my ($RmIfStale,$RmIfDirStale,
            $Report,$Count,$PatternName,$PostCQ);
        my $pattern;
        my $patternMatches = 0;
        for $pattern (sort keys %filePatterns) {
            if ($file =~ /$pattern/) {
                my @attributes = split(",", $filePatterns{$pattern});
                $patternMatches++;
                if ($outputLevel >= 3) {
                    my $ageHr = sprintf("%.1f", ($age / 60 / 60));
                    $fileBuffer .= $tab2. 
                        msg(["TCM_T_PATTERN_MATCH", $pattern,
                             join(",",@attributes), $ageHr])."\n";
                }
                for (@attributes) {
                    my ($attName,$attValue) = split("=",$_);

                    # Determine rule name                   

                    if ($attName eq "Rule") {
                        $PatternName = $attValue;

                    # Count - Increment the indicated counter if found
                    } elsif ($attName eq "Count") {
                        if (defined $sessionStats{$attValue}) {
                            $sessionStats{$attValue}++;
                        } else {
                            $fileBuffer .= 
                               $tab2.msg(["TCM_T_UNKNOWN_COUNT",
                                         $attName,$attValue])."\n";
                        }

                    # PostCQ - Note unposted batch found, possibly stale 
                    } elsif ($attName eq "PostCQ") {
                        $unpostedCQbatch = 1;
                        $batchFile = $path;
                        if ($age > $postOverAgeSecs) {
                            $postCQbatch = 1;
                            $sessionStats{"FoundCQBatchStale"}++;
                        }

                    # RmIfStale - Note file can be removed if over specific age
                    } elsif ($attName eq "RmIfStale") {
                        $RmIfStale = 1;
                        if ($age > $maxAgeSecs) {
                            $deleteFile .= msg(["TCM_T_RULE_OVER_AGE",
                                "RmIfStale", &showTime($age,"hours"), 
                                &showTime($maxAgeSecs,"hours")]);
                        }

                    # RmIfDirStale - Note file can be removed if over specific age                    
                    } elsif ($attName eq "RmIfDirStale") {
                        $RmIfDirStale = 1;

                    # Report - report file as interesting
                    } elsif ($attName eq "Report") {
                        $Report = 1;
                        $fileBuffer .= 
                            $tab2.msg(["TCM_T_INTERESTING_FILE"])."\n";
                    } else {
                        $fileBuffer .= 
                            $tab2.msg(["TCM_T_UNKNOWN_FILE_PATTERN",
                                      $attName,$attValue])."\n";
                    }
                }
            }
        }
        if ($patternMatches == 0) {
            $fileBuffer .= $tab2.msg(["TCM_T_FILE_UNEXPECTED"])."\n";
        }

        ########################################
        # Apply accumulated file pattern rules

        if (! defined $RmIfDirStale) {
            push @ReasonsNotToRmDir, $file;
        }

        if ($deleteFile ne "") {
            if (&GetParm("CM_NOEXEC") eq "1") {
                if ($outputLevel >= 2) {
                    $fileBuffer .= $tab2.msg(["TCM_T_WOULD_DELETE_FILE"]).
                                   "$deleteFile\n";
                }
            } else {
                unlink($path);
                if (-e $path) {
                    $fileBuffer .= $tab2.msg(["TCM_T_CANT_DELETE_FILE"])."\n";
                    $errorCount++;
                } else {
                    $sessionStats{"DeletedFiles"}++;
                    if ($outputLevel >= 2) {
                        $fileBuffer .= $tab2.msg(["TCM_T_DELETED_FILE",
                                                 $deleteFile])."\n";
                    }
                }
            }
        }
        if ($fileBuffer ne "") {
            $dirBuffer .= $tab1."$file\n$fileBuffer";
        }
    }

    ##################################################
    # Post batches using a subprocess command so that errors in 
    # the posting process don't kill the main cache manager process.

    if ($postCQbatch eq "1" && &GetParm("CM_NOEXEC") eq "1") {
        logMessage(1,msg(["TCM_T_WOULD_POST_BATCH", $batchFile])."\n"); 

    } elsif ($postCQbatch eq "1") {
        my $cmd = &commandLine();
        $batchFile =~ s/\\/\\\\/g;
        if ($batchFile =~ /\s/) {
            $batchFile = "\"$batchFile\"";
        }
        # Note: Filtering out -p TriggerInterface,TriggerCacherMgr
        # might make sense to allow local user subclasses to do
        # actual posting but could also enable user interaction 
        # and require CQ login, etc. so avoid for now. 

        $cmd =~ s/\-op manage[\s\S]*$/-op batch do,$batchFile/;

        logMessage(1,">>>>>>>>>> Post Batch ($sessionDir)\n");
        logMessage(1,msg(["TCM_T_POSTING_BATCH", $cmd])."\n");
        $ENV{"CQCC_CACHE_MGR_SUBPROCESS"} = 1;
        my $result = system($cmd); 
        $ENV{"CQCC_CACHE_MGR_SUBPROCESS"} = undef;
        $sessionStats{"PostedCQBatchStale"}++;
        if ($result ne 0) {
            &IOMessage("ERROR", ["TCM_E_FAILED_POST",$batchFile]);
            $sessionStats{"PostedCQBatchStaleFail"}++;
            $errorCount++;
        } else {
            $sessionStats{"PostedCQBatchStaleOK"}++;
        }
        logMessage(1,"<<<<<<<<< Post Batch - End\n");
    }

    ##################################################
    # Delete directory if youngest file is over stale Age
    # and all files have RmIfDirStale property

    if (($#files == -1) ||
        ($youngestAge >= $maxDirAgeSecs && $#ReasonsNotToRmDir == -1)) {
        if (&GetParm("CM_NOEXEC") eq "1") {
            my $ageHr = sprintf("%.1f", ($maxDirAgeSecs / 60 / 60));
            if ($outputLevel >= 1) {
                $dirBuffer .= $tab1.msg(["TCM_T_WOULD_DELETE_DIR",$ageHr])."\n";
            }
        } else {
            @files = dirList($dir);
            for $file (@files) {
                my $path = $dir . $ps . $file;
                unlink($path);
                if (-e $path) {
                    $dirBuffer .= $tab1.msg(["TCM_T_CANT_DELETE_FILE",$path])."\n";
                    $errorCount++;
                } else {
                    $sessionStats{"DeletedFiles"}++;
                    if ($outputLevel >= 2) {
                        $dirBuffer .= $tab1."$file\n" . 
                            $tab2 . msg(["TCM_T_DELETED_FILE",
                                         "Rule: RmIfDirStale"])."\n";
                    }
                }
            }
            my @files2 = dirList($dir);
            if ($#files2 != -1) {
                $dirBuffer .= $tab2.msg(["TCM_T_CANT_DELETE_DIR"])."\n";
                $errorCount++;
            } else {
                my ($status,$output) = 
                     &deleteDir($dir,$tab1,"DeletedSessionDirs",
                                "RmDirIfStale");
                $errorCount += ($status ? 0 : 1);
                $dirBuffer .= $output;
            }
        }
    }

    if ($dirBuffer ne "") {
        logMessage(0,"\n$sessionDir\n$dirBuffer");
    }

    ##################################################
    # Post session stats to run stats
    for (sort keys %sessionStats) {
        if ($sessionStats{$_} > 0) {
            $stats{$_} = $stats{$_} + $sessionStats{$_};
        }
    }
    $stats{"TotalSessions"}++;

    return $errorCount;
}

######################################################################
#::method processCQlogin

=over 6

=item C<processCQlogin>

=back

=cut

######################################################################

sub processCQlogin {
    my ($self) = @_;

    ########################################
    # If CQLogin information is provided then
    # decrypt the password and pass information 

    my $cqLogin = &GetParm("CM_CQLOGIN");
    if ($cqLogin ne "") { 
        my ($user,$pass) = split(":",$cqLogin,2);
        my $cqResult = 1;
        if ($] ge "5.006") {
            eval { 
                require CQCCext;
                eval { 
                    $pass = CQCCext::decrypt_password($pass);
                };
                if ($@) {
                    $cqResult = 0;
                }
            };
            if ($@) {
                $cqResult = 0;
            }
        } else {
            &IOMessage("FATAL", ["TCM_E_CQLOGIN_UNSUPPORTED"]);
        }
        if ($cqResult != 1) {
            &IOMessage("FATAL", ["TCM_E_CQLOGIN_FAILED"]);
        }
        if (defined $pass) {
            $ENV{"CQCC_TEST_USER"} = $user;
            $ENV{"CQCC_TEST_PW"} = $pass;
        }
        # Signal CQ not to prompt for login
        $ENV{"CQCC_CQ_NOLOGIN"} = 1;
    }
}

######################################################################
#::function logMessage 

=over 6

=item C<logMessage(level,message)>

This function logs output messages to the current output log. 

=back

=cut

######################################################################

sub logMessage {
    my ($level,$message) = @_;
    if ($level <= $outputLevel) {
        &IOPrint($message);
    }
}

######################################################################
#::function dirList 

=over 6

=item C<dirList(pathList)>

This utility function takes a list of directory pathname components,
combines them and returns a list of the contents of the directory
after stripping out the current and parent directories.

=back

=cut

######################################################################

sub dirList {
    my (@dirComponents) = @_;
    my $path = join($ps,@dirComponents);

    opendir(DIRHANDLE,$path);
    my @files = readdir(DIRHANDLE);
    closedir(DIRHANDLE);
    
    my @fileList;
    for (@files) {
        if ($_ ne "." && $_ ne "..") {
            push @fileList, $_;
        }
    }
    return sort @fileList;
}

######################################################################
#::function commandLine

=over 6

=item C<commandLine>

This utility function creates a copy of the command line that 
can be used to invoke subprocesses with the same command line. 

=back

=cut

######################################################################

sub commandLine {
    my %replacements = @_;
    my $perl = &shortFileName($^X);
    my $program = &shortFileName($0);
    my @args;
    my $arg;    

    my $inc;
    # Determine home directory for CQCC from @INC path
    for (@INC) {
        if (-d "$_$ps"."CQCC") {
            $inc = "-I" . &shortFileName($_);
            last;
        }
    }

    # For Unix, check to see if we are running cqperl based on @INC contents
    # If it is, assume cqcc_launch did the cq setup and prepped the path to
    # find cqperl easily. 

    if ($windows eq 0) {
        for (@INC) {
            if ($_ =~ /clearquest/) {
                $perl = "cqperl";
                last;
            }
        }
    }

    for $arg (@ARGV) {
        if ($replacements{$arg}) {
            $arg = $replacements{$arg};
        }
        if ($arg =~ / /) {
            $arg = "\"$arg\"";
        }
        push @args,$arg;
    }
    if ($ENV{"CQCC_LAUNCH_DEBUG"} > 1) {
        $perl .= " -d";
    }
    my $cmd = "$perl $inc $program " . join(" ",@args);
    return $cmd;
}

######################################################################
#::function shortFileName

=over 6

=item C<value = shortFileName(longName)>

This utility function returns the Windows-friendly short name
of a given file or directory name. If the name contains spaces,
it will be quoted as necessary.

=back

=cut

######################################################################

sub shortFileName {
    my ($longName) = @_;
    if ($windows) {
        eval {
            require Win32;
            my $shortName = Win32::GetShortPathName($longName);
            if ($shortName =~ / /) {
                $shortName = "\"" . $shortName . "\"";
            }
            $longName = $shortName;
        }
    }
    return $longName;
}

######################################################################
#::function showTime

=over 6

=item C<value = showTime(seconds,format)>

This utility function converts seconds into either "mins" or "hours"
rounding to the nearest minute or 0.1 hours.

=back

=cut

######################################################################

sub showTime {
    my ($secs,$format) = @_;
    my $result;
    if ($format eq "mins") {
        $result = sprintf("%d", ($secs+30)/60);
    } elsif ($format eq "hours") {
        $result = sprintf("%.1f", $secs/3600);
    } else {
        $result = "(ERROR: showTime given unexpected format $format)";
    }
    return $result;
}

######################################################################
#::function deleteDir

=over 6

=item C<(success,output) = deleteDir(dir, tab, stat, rule)>

This utility function deletes a given directory, reports an error
if unsuccessful, and increments the indicated statistic when successful.

It returns a list containing a status (1 if successful, 0 if not)
and an output string for logging using the tab value for indentation
and showing the rule justifying removing the directory.

=back

=cut

######################################################################

sub deleteDir {
    my ($dir,$tab,$stat,$rule) = @_;
    my $output = "";
    my $status = 1;

    rmdir($dir);
    if (-d $dir) {
        $output = $tab.msg(["TCM_T_CANT_DELETE_DIR2",$dir])."\n";
        $status = 0;        

    } elsif ($outputLevel >= 1) {
        $output = $tab.msg(["TCM_T_DELETED_DIR",$rule])."\n";
    }
    if (defined $sessionStats{$stat}) {
        $sessionStats{$stat}++;
    } elsif (defined $stats{$stat}) {
        $stats{$stat}++;
    }
    return ($status,$output);
}

######################################################################
#::function msg 

=over 6

=item C<value = msg(msgTag)>

This utility function accesses the Message catalog to retrieve a
translated message value or component. 

=back

=cut

######################################################################

sub msg {
    my $msgCat = &IOstd()->{MSGCAT};
    return $msgCat->Message(@_);
}

######################################################################
#::Tail
######################################################################

1;
